/*
 * $Id: s.gscdd.c 1.10 01/06/09 10:07:35-07:00 mjacob@bird.feral.com $
 * Copyright (c) 1996, 1997 by Matthew Jacob
 * Copyright (c) 2001,2011 by Frank Zago
 * Copyright (c) 2011 Gresham Enterprise Storage
 *
 *	This software is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU Library General Public
 *	License as published by the Free Software Foundation; version 2.
 *
 *	This software is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *	Library General Public License for more details.
 *
 *	You should have received a copy of the GNU Library General Public
 *	License along with this software; if not, write to the Free
 *	Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *	The author may be reached via electronic communications at
 *
 *		mjacob@feral.com
 *
 *	or, via United States Postal Address
 *
 *		Matthew Jacob
 *		1831 Castro Street
 *		San Francisco, CA, 94131
 */

#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/errno.h>
#include <sys/sysmacros.h>
#include <sys/syspest.h>
#include <sys/ioctl.h>
#include <sys/i_machine.h>
#include <sys/systm.h>
#include <sys/conf.h>
#include <sys/devinfo.h>
#include <sys/lockl.h>
#include <sys/device.h>
#include <sys/uio.h>
#include <sys/watchdog.h>
#include <sys/errids.h>
#include <sys/trchkid.h>
#include <sys/priv.h>
#include <sys/iostat.h>
#include <sys/bootrecord.h>
#include <sys/scsi.h>
#ifdef FCP
#include <sys/scsi_buf.h>
#endif
#include <sys/malloc.h>
#include <sys/sleep.h>
#include <sys/fp_io.h>
#include <sys/pin.h>
#include <sys/lock_alloc.h>
#include <sys/err_rec.h>

#define	DD_LOCK	37

#include "gscdds.h"
#include "gscdds_internal.h"

/*
 * Local Definitions
 */

#define	HKWD_GSC_DD	0x66600000

#define	_COM_TRACE(b, var)	\
	var = strlen(b); trcgenk(0, HKWD_GSC_DD, var, var, b)

#define	Trace0(val, str)	\
	if (scudebug >= val) { \
		int icxq; \
		_COM_TRACE(str, icxq); \
	}
#define	Trace1(val, fmt, arg1)	\
	if (scudebug >= val) { \
		int icxq; char buf[256]; \
		(void) sprintf(buf, fmt, arg1); \
		_COM_TRACE(buf, icxq); \
	}
#define	Trace2(val, fmt, arg1, arg2)	\
	if (scudebug >= val) { \
		int icxq; char buf[256]; \
		(void) sprintf(buf, fmt, arg1, arg2); \
		_COM_TRACE(buf, icxq); \
	}
#define	Trace3(val, fmt, arg1, arg2, arg3)	\
	if (scudebug >= val) { \
		int icxq; char buf[256]; \
		(void) sprintf(buf, fmt, arg1, arg2, arg3); \
		_COM_TRACE(buf, icxq); \
	}
#define	Trace4(val, fmt, arg1, arg2, arg3, arg4)	\
	if (scudebug >= val) { \
		int icxq; char buf[256]; \
		(void) sprintf(buf, fmt, arg1, arg2, arg3, arg4); \
		_COM_TRACE(buf, icxq); \
	}

#define	Trace5(val, fmt, arg1, arg2, arg3, arg4, arg5)	\
	if (scudebug >= val) { \
		int icxq; char buf[256]; \
		(void) sprintf(buf, fmt, arg1, arg2, arg3, arg4, arg5); \
		_COM_TRACE(buf, icxq); \
	}

#define	Trace6(val, fmt, arg1, arg2, arg3, arg4, arg5, arg6)	\
	if (scudebug >= val) { \
		int icxq; char buf[256]; \
		(void) sprintf(buf, fmt, arg1, arg2, arg3, arg4, arg5, arg6); \
		_COM_TRACE(buf, icxq); \
	}

typedef struct {
#ifdef FCP
    struct scsi_buf scsibuf;
#else
    struct sc_buf scsibuf;
#endif
    uint index;
} gsc_buf_t;

typedef struct {
    Simple_lock dd_lock;
    Simple_lock buf_lock;
    struct file *fp;		/* file pointer */
    gsc_buf_t cbuf;
#define	cmdbuf cbuf.scsibuf	/* buffer for command */
    gsc_buf_t rbuf;
#define rqsbuf rbuf.scsibuf	/* buffer for request sense */
    dev_t dev;			/* Adapter dev */
    unsigned long long scsi_id;	/* target SCSI ID */
    unsigned long long lun_id;	/* logical unit */
    u_char isopen;		/* device is open */
    u_char iscfg;		/* non-zero to show this as configured */
    u_char unstart;		/* stop device on unconfigure */
    u_char needresume;		/* needs an SC_RESUME with next command */
    char devname[ERR_NAMESIZE]; /* device name (eg. gsc0) */
	char ww_name[20];			/* 0x followed by 16 chars */
	char node_name[20];			/* 0x followed by 16 chars */
} gsc_softc_t;

/* Try to offset the differences between the SCSI and FCP APIs */
#ifdef FCP
#define SCSICMD_SHIFT 1	/* no lun field */
#else
#define SCSICMD_SHIFT 0	/* has lun field */
#define adapter_status general_card_status
#endif

/*
 * Device Driver Entry Points
 */
int gsc_config(dev_t, int, struct uio *);
static int gsc_open(dev_t);
static int gsc_close(dev_t);
static int gsc_ioctl(dev_t, int, void *, ulong);
static void gscdd_intr(struct buf *);

/*
 * Static Data
 */
static int scudebug = 10;
static int nunits = 0;
static gsc_softc_t softinfo[MAX_UNITS];
lock_t config_lock = { LOCK_AVAIL };


/*
 * Local Function Prototypes
 */

static int gsopen(gsc_softc_t *);
static void gsclose(gsc_softc_t *, dev_t);
static int gsccmd(gsc_softc_t *sp, scmd_t *, ulong);
static int make_rqs(gsc_softc_t *, char, char *, int, int);
static int pin_and_attach (struct buf *, int);
static int unpin_and_detach (struct buf *);

/*
 * Configuration Routines
 */

int
gsc_config(dev_t devno, int cmd, struct uio * uiop)
{
    struct gsc_ddsinfo ddsinfo;
#ifdef FCP
	struct gsc_qvpd qvpd;
#endif
    gsc_softc_t *sp;
    int result, unit;

    static struct devsw gsc_dsw = {
		gsc_open,	/* entry point for open routine */
		gsc_close,	/* entry point for close routine */
		nodev,		/* entry point for read routine */
		nodev,		/* entry point for write routine */
		gsc_ioctl,	/* entry point for ioctl routine */
		nodev,		/* entry point for strategy routine */
		0,		/* pointer to tty device structure */
		nodev,		/* entry point for select routine */
		gsc_config,	/* entry point for config routine */
		nodev,		/* entry point for print routine */
		nodev,		/* entry point for dump routine */
		nodev,		/* entry point for mpx routine */
		nodev,		/* entry point for revoke routine */
		NULL,		/* pointer to device specific data */
		NULL,		/* select pointer */
#ifdef __64BIT__
		DEV_64L |
#endif
		DEV_MPSAFE
    };

    Trace2(2, "CFG_INIT: devno=%lx cmd=%d", (long)devno, cmd);
	
    if (lockl(&config_lock, LOCK_SHORT) != LOCK_SUCC) {
		return (EINVAL);
    }
    unit = minor(devno);
    if (unit < 0 || unit >= MAX_UNITS) {
		Trace2(0, "%d: bad unit %d", __LINE__, unit);
		result = EINVAL;
		unlockl(&config_lock);
		return (result);
    }
	
    switch (cmd) {
    case CFG_INIT:
		Trace2(2, "CFG_INIT: unit %d nunit %d", unit, nunits);

		/*
		 * Initialize softinfo, first time around.
		 */
		if (nunits == 0) {
			bzero(softinfo, sizeof (softinfo));
		}
		/*
		 * Copy in DDS information
		 */
		uiomove((caddr_t) &ddsinfo, sizeof ddsinfo, UIO_WRITE, uiop);

		Trace3(2, "CFG_INIT: busid=%x, scsi_id=0x%llx, lun=0x%llx", ddsinfo.busid, ddsinfo.scsi_id, ddsinfo.lun_id);

		sp = &softinfo[unit];
		if (sp->iscfg) {
			Trace1(0, "CFG_INIT: unit %d already configd", unit);
			result = EBUSY;
			break;
		}
		lock_alloc(&sp->dd_lock, LOCK_ALLOC_PIN, DD_LOCK, -1);
		lock_alloc(&sp->buf_lock, LOCK_ALLOC_PIN, DD_LOCK, -1);
		simple_lock_init(&sp->dd_lock);
		simple_lock_init(&sp->buf_lock);
		sp->dev = ddsinfo.busid;
		sp->scsi_id = ddsinfo.scsi_id;
		sp->lun_id = ddsinfo.lun_id;
		sp->cbuf.index = sp->rbuf.index = unit;
		strncpy(sp->devname, ddsinfo.devname, ERR_NAMESIZE);
		sp->ww_name[0] = 0;
		sp->node_name[0] = 0;

		/*
		 * If this is the first time through:
		 *   Add entry to the device switch table to call this driver
		 *   Pin driver code.
		 */
		if (nunits == 0) {
			result = devswadd(devno, &gsc_dsw);
			if (result != 0) {
				Trace1(0, "CFG_INIT: devswadd result: %d", result);
				break;
			}
			result = pincode((int (*) ()) gscdd_intr);
			if (result) {
				Trace1(0, "CFG_INIT: pincode result: %d", result);
				devswdel(devno);
				break;
			}
		}
		sp->iscfg = 1;
		result = gsopen(sp);
		if (result) {
			Trace2(0, "CFG_INIT: gsopen returns %d for unit %d", result, unit);
			sp->iscfg = 0;
			gsclose(sp, devno);
			break;
		}
		if (nunits <= unit)
			nunits = unit + 1;
		sp->iscfg = 1;

		Trace1(0, "CFG_INIT: device %s is available", sp->devname);
		break;
		
    case CFG_TERM:
		Trace1(2, "CFG_TERM unit %d", unit);
		result = 0;
		sp = &softinfo[unit];
		if (sp->iscfg == 0) {
			Trace1(0, "CFG_TERM: unit %d not already configd", unit);
			result = ENXIO;
			break;
		} else if (sp->isopen) {
			Trace1(0, "CFG_TERM: unit %d open", unit);
			result = EBUSY;
			break;
		}
		sp->iscfg = 0;	/* block further actions */
		gsclose(sp, devno);
		break;

#ifdef FCP
	case CFG_QVPD:
		Trace1(2, "CFG_QVPD unit %d", unit);
		result = 0;
		sp = &softinfo[unit];
		if (sp->iscfg == 0) {
			Trace1(0, "CFG_TERM: unit %d not configd", unit);
			result = ENXIO;
			break;
		}

		strcpy(qvpd.node_name, sp->node_name);
		strcpy(qvpd.ww_name, sp->ww_name);

		uiomove((caddr_t) &qvpd, sizeof qvpd, UIO_READ, uiop);

		break;
#endif
		
    default:
		result = EINVAL;
		break;
    }
    unlockl(&config_lock);
    return (result);
}

/*
 * Validate that devno is indeed for a SCSI adapter, and set up stuff for it.
 */
static int
gsopen(gsc_softc_t * sp)
{
    struct file *fp;
    int r;
    struct devinfo di;
#ifdef FCP
    struct scsi_sciolst sciolst;
#endif
	
    Trace2(2, "gsopen: %d.%d", major(sp->dev), minor(sp->dev));
    sp->fp = NULL;
    r = fp_opendev(DEV32TODEV(sp->dev), DREAD|DWRITE|DKERNEL, NULL, 0, &fp);
    if (r) {
		Trace3(0, "%d: fp_opendev unit %d=%d", __LINE__, sp->cbuf.index, r);
		return (r);
    }
    r = fp_ioctl(fp, IOCINFO, (caddr_t) &di, 0);
    if (r) {
		Trace3(0, "%d: fp_ioctl unit %d=%d", __LINE__, sp->cbuf.index, r);
		(void) fp_close(fp);
		return (r);
    }

#ifdef FCP
    if (di.devtype != DD_BUS || di.devsubtype != DS_FCP) {
		Trace4(0, "%d: not FCP bus on unit %d (%d, %d)", __LINE__, sp->cbuf.index, di.devtype, di.devsubtype);
		(void) fp_close(fp);
		return EINVAL;
    }
#else
    if (di.devtype != DD_BUS || di.devsubtype != DS_SCSI) {
		Trace4(0, "%d: not SCSI bus on unit %d (%d, %d)", __LINE__, sp->cbuf.index, di.devtype, di.devsubtype);
		(void) fp_close(fp);
		return EINVAL;
    }
#endif

    sp->fp = fp;
    sp->unstart = 1;

#ifdef FCP
    memset(&sciolst, 0, sizeof(struct scsi_sciolst));
    sciolst.version = SCSI_VERSION_1;
    sciolst.flags = ISSUE_LOGIN;
    sciolst.scsi_id = sp->scsi_id;
    sciolst.lun_id = sp->lun_id;
	r = fp_ioctl(sp->fp, SCIOLSTART, (caddr_t)&sciolst, 0);
	if (r) {
		Trace2(0, "%d: SCIOLSTART failed (error %d)", __LINE__, r);
		Trace2(2, "scsi_id=%llx, lun_id=%lld", sp->scsi_id, sp->lun_id);

		sp->unstart = 0;
		return r;
    }

	if (sciolst.adap_set_flags & WWN_VALID) {
		Trace2(0, "%d: World Wide Name is %016llx", __LINE__, sciolst.world_wide_name);
		Trace2(0, "%d: Node name is       %016llx", __LINE__, sciolst.node_name);

		sprintf(sp->node_name, "%016llx", sciolst.node_name);
		sprintf(sp->ww_name, "%016llx", sciolst.world_wide_name);
	}
#else
	if (fp_ioctl(sp->fp, SCIOSTART, (caddr_t) IDLUN(sp->scsi_id, sp->lun_id), 0)) {
		sp->unstart = 0;
    }
#endif

    return (0);
}

/*
 * Shut down a device
 */
static void
gsclose(gsc_softc_t *sp, dev_t devno)
{
    int i;
#ifdef FCP
	int r;
    struct scsi_sciolst sciolst;
#endif

    if (sp->fp != NULL && sp->unstart) {
#ifdef FCP
		memset(&sciolst, 0, sizeof(struct scsi_sciolst));
		sciolst.version = SCSI_VERSION_1;
		sciolst.scsi_id = sp->scsi_id;
		sciolst.lun_id = sp->lun_id;
		r = fp_ioctl(sp->fp, SCIOLSTOP, (caddr_t)&sciolst, 0);
		if (r)
			Trace2(0, "%d: SCIOLSTOP failed (error %d)", __LINE__, r);
#else
		(void) fp_ioctl(sp->fp, SCIOSTOP, (caddr_t) IDLUN(sp->scsi_id, sp->lun_id), 0);
#endif
		sp->unstart = 0;
    }
    if (sp->fp) {
		(void) fp_close(sp->fp);
		sp->fp = NULL;
    }
    for (i = 0; i < MAX_UNITS; i++) {
		if (softinfo[i].iscfg) {
			Trace1(0, "gsclose: unit %d still confd", i);
			break;
		}
    }
	lock_free(&sp->dd_lock);
	lock_free(&sp->buf_lock);
    if (i == MAX_UNITS) {
		Trace0(0, "gsclose: All unconfigured now");
		devswdel(devno);
		unpincode((int (*) ()) gscdd_intr);
    }
}

/*
 * VFS entry points
 */

static int
gsc_open(dev_t devno)
{
    gsc_softc_t *sp;
    int unit = minor(devno);
	
    Trace1(2, "gsc_open: open unit %d", unit);
    if (unit < 0 || unit >= MAX_UNITS) {
		return (ENODEV);
    }
    sp = &softinfo[unit];
    if (sp->iscfg == 0 || sp->fp == NULL) {
		Trace2(0, "%d: bad unit (%d)", __LINE__, unit);
		return (ENODEV);
    }
    simple_lock(&sp->dd_lock);
    if (sp->isopen) {
		simple_unlock(&sp->dd_lock);
		return (EBUSY);
    }
    sp->isopen = 1;
    simple_unlock(&sp->dd_lock);
    return (0);
}

static int
gsc_close(dev_t dev)
{
    gsc_softc_t *sp;
    int unit = minor(dev);
	
    Trace1(2, "gsc_close: close unit %d", unit);
    if (unit < 0 || unit >= MAX_UNITS) {
		return (ENODEV);
    }
    sp = &softinfo[unit];
    if (sp->iscfg == 0) {
		return (ENODEV);
    }
    simple_lock(&sp->dd_lock);
    if (!sp->isopen) {
		simple_unlock(&sp->dd_lock);
		return (ENOENT);
    }
    sp->isopen = 0;
    simple_unlock(&sp->dd_lock);
    return (0);
}

static int
gsc_ioctl(dev_t dev, int cmd, void *arg, ulong dflag)
{
    gsc_softc_t *sp;
    int unit = minor(dev);
	int rc = 0;
	
    if (unit < 0 || unit >= MAX_UNITS) {
		return (ENODEV);
    }
    sp = &softinfo[unit];
    if (sp->iscfg == 0 || sp->fp == NULL) {
		Trace2(0, "gsc_ioctl: bad unit %d (cfg=%d)", unit, sp->iscfg);
		return (ENODEV);
    }

    simple_lock(&sp->dd_lock);

    if (!sp->isopen) {
		simple_unlock(&sp->dd_lock);
		return (ENOENT);
    }

    Trace1(2, "gsc_ioctl: command %08x", cmd);

    switch (cmd) {
    case GSC_CMD: {
		/* handle 32 bit ioctl on 32 kernel and 64 bits ioctl on 64
		 * bits kernel. */
		scmd_t scmd;
		scmd_t *scmdptr;
		
		if (dflag & DKERNEL) {
			scmdptr = (scmd_t *)arg;
		} else {
			rc = copyin((caddr_t) arg, (caddr_t) &scmd, sizeof (scmd_t));

			if (rc) {
				Trace2(0, "%d: copyin=%d", __LINE__, rc);
			}
			scmdptr = &scmd;
		}

		if (!rc) {
			rc = gsccmd(sp, scmdptr, dflag);
		}

		if (!rc) {
			if ((dflag & DKERNEL) == 0) {
				rc = copyout((caddr_t) &scmd, arg, sizeof (scmd_t));
			}
		}
	}
	break;

#ifdef __64BIT__
	case GSC_CMD32: {
		/* Process 32 bits ioctl on 64 bits kernel. */
		scmd_t scmd;
		scmd32_t scmd32;
		
		rc = copyin((caddr_t) arg, (caddr_t) &scmd32, sizeof (scmd32_t));
		
		if (rc) {
			Trace2(0, "%d: copyin=%d", __LINE__, rc);
		}

		if (!rc) {
			scmd.cdb = (char *)(uintptr_t)scmd32.cdb;
			scmd.cdblen = scmd32.cdblen;
			scmd.data_buf = (char *)(uintptr_t)scmd32.data_buf;
			scmd.datalen = scmd32.datalen;
			scmd.sense_buf = (char *)(uintptr_t)scmd32.sense_buf;
			scmd.senselen = scmd32.senselen;
			scmd.statusp = (char *)(uintptr_t)scmd32.statusp;
			scmd.rw = scmd32.rw;
			scmd.timeval = scmd32.timeval;
			
			rc = gsccmd(sp, &scmd, dflag);
		}
		
		if (!rc) {
			/* datalen and senselen may have been modified. All other
			 * field are read-only. */ 
			scmd32.datalen = scmd.datalen;
			scmd32.senselen = scmd.senselen;

			rc = copyout((caddr_t) &scmd32, arg, sizeof (scmd32_t));
		}
	}
	break;
#endif
		
    case GSC_SETDBG: {
		int i;
		rc = copyin(arg, (caddr_t) &i, sizeof (int));
		if (!rc) {
			scudebug = i;
			rc = copyout((caddr_t) &scudebug, arg, sizeof (int));
		}
	}
	break;

    default:
		rc = ENOTTY;
    }

	simple_unlock(&sp->dd_lock);

    Trace1(2, "gsc_ioctl: exit with %d", rc);

	return(rc);
}


/****************************************************************************/

static int
gsccmd(gsc_softc_t *sp, scmd_t *l, ulong dflag)
{
    char sbyte, albits;
#ifdef FCP
	struct scsi_buf *usc;
#else
    struct sc_buf *usc;
#endif
    struct buf *Ubp;
    int actual_data, actual_sense, r, r2, ival, upin, rqvalid, once;
	
    Trace6(1, "%d: cdblen=%d datalen=%d snslen=%d rw=%d tv=%d", __LINE__,
		   l->cdblen, l->datalen, l->senselen, l->rw, l->timeval);

    sbyte = 0;
    rqvalid = upin = r = r2 = 0;
    usc = &sp->cmdbuf;
    Ubp = &usc->bufstruct;
    bzero(usc, sizeof (*usc));

    actual_data  = 0;
    actual_sense = 0;
	
    /*
     * Check some parameters...
     */
#ifndef FCP
	/* Only FCP API supports 16 bytes CDBs. */
	if (l->cdblen == 16) {
		r = EINVAL;
		goto out;
	}
#endif

    if (l->cdblen != 6 && l->cdblen != 10 && l->cdblen != 12 && l->cdblen != 16) {
		r = EINVAL;
		goto out;
    }
	
    /*
     * Setup sc_buf structure
     */
    Ubp->b_iodone = gscdd_intr;
    Ubp->b_dev = DEV32TODEV(sp->dev);
    Ubp->b_flags = B_BUSY | B_MPSAFE;
    Ubp->b_resid = Ubp->b_bcount = l->datalen;
    Ubp->b_xmemd.aspace_id = XMEM_INVAL;
    Ubp->b_event = EVENT_NULL;
	
    if (l->datalen) {
		Ubp->b_un.b_addr = l->data_buf;
		if (l->rw) {
			Ubp->b_flags |= B_READ;
		}
		if (dflag & DKERNEL) {
			r = pin_and_attach(Ubp, UIO_SYSSPACE);
		} else {
			r = pin_and_attach(Ubp, UIO_USERSPACE);
		}
		if (r) {
			Trace2(0, "%d: pinu buf %d", __LINE__, r);
			goto out;
		}
		upin = 1;
    }
#ifdef FCP
	usc->scsi_id = sp->scsi_id;
	usc->version = SCSI_VERSION_2;
#else
    usc->scsi_command.scsi_id = sp->scsi_id;
#endif
    usc->scsi_command.scsi_length = l->cdblen;
    if (dflag & DKERNEL) {
		bcopy(l->cdb, (caddr_t)&usc->scsi_command.scsi_cmd, l->cdblen);
    } else {
		r = copyin(l->cdb, (caddr_t) & usc->scsi_command.scsi_cmd, l->cdblen);
		if (r != 0) {
			goto out;
		}
    }

    /* Setting lun. */
#ifdef FCP
    usc->lun_id = sp->lun_id;
	albits = 0;					/* unused */
#else
    /* Setting lun in SCSI CDB as well as sc_buf structure */
    usc->lun = sp->lun_id;
    usc->scsi_command.scsi_cmd.lun &= 0x1F;
    usc->scsi_command.scsi_cmd.lun |= (sp->lun_id << 5) & 0xE0;
    albits = usc->scsi_command.scsi_cmd.lun;
#endif

    usc->timeout_value = l->timeval;
    if (sp->needresume) {
		usc->flags |= SC_RESUME;
		sp->needresume = 0;
    }
	
    if (scudebug > 1) {
		unsigned char *c = (unsigned char *) &usc->scsi_command.scsi_cmd;
		char cdbuf[64];
		(void) sprintf(cdbuf,
				"0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x "
				"0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x",
				c[0], c[1], c[2], c[3], c[4], c[5],
				c[6], c[7], c[8], c[9], c[10], c[11]);
		Trace2(0, "%d: cdb=%s", __LINE__, cdbuf);
    }
	
    once = 0;
 again:
    Ubp->b_flags &= ~B_DONE;
    r = devstrat(Ubp);
    if (r == 0) {
		ival = disable_lock(INTCLASS1, &sp->buf_lock);
		while ((Ubp->b_flags & B_DONE) == 0) {
			e_sleep_thread(&Ubp->b_event, &sp->buf_lock, LOCK_HANDLER);
		}
		Trace0(1, "gsccmd: B_DONE set");
		unlock_enable(ival, &sp->buf_lock);
    } else {
		/*
		 * If ENXIO,  We never actually got started.
		 */
		if (r == ENXIO && once == 0) {
			once++;
			usc->flags |= SC_RESUME|SC_DELAY_CMD;
			goto again;
		}
		sp->needresume = 1;
		Trace2(1, "%d: devstrat=%d", __LINE__, r);
		goto out;
    }
    actual_data = l->datalen - Ubp->b_resid;
    Trace4(1, "%d: b_flags %lx b_error %d b_resid %ld", __LINE__,
		   (long)Ubp->b_flags, Ubp->b_error, (long)Ubp->b_resid);
    Trace5(1, "%d: sv %x st %x gc %x as %x", __LINE__,
		   usc->status_validity, usc->scsi_status,
		   usc->adapter_status, usc->adap_q_status);
	
    if (Ubp->b_flags & B_ERROR) {
		r = Ubp->b_error;
		sp->needresume = 1;
    }
	
    if (usc->status_validity & SC_SCSI_ERROR) {
		sbyte = (usc->scsi_status & SCSI_STATUS_MASK);
		sp->needresume = 1;
		if (sbyte == SC_CHECK_CONDITION && l->senselen) {
#ifdef FCP
			struct scsi_buf *usl;
#else
			struct sc_buf *usl;
#endif
			struct buf *Sbp;
			
			r = make_rqs(sp, albits, l->sense_buf, l->senselen,
						 (dflag & DKERNEL) != 0);
			if (r) {
				Trace2(0, "%d: make_rqs=%d", __LINE__, r);
				goto out;
			}
			usl = &sp->rqsbuf;
			Sbp = &usl->bufstruct;
			r = devstrat(Sbp);
			if (r == 0) {
				ival = disable_lock(INTCLASS1, &sp->buf_lock);
				while ((Sbp->b_flags & B_DONE) == 0) {
					e_sleep_thread(&Sbp->b_event, &sp->buf_lock, LOCK_HANDLER);
				}
				unlock_enable(ival, &sp->buf_lock);
			} else {
				Trace2(0, "%d:ds=%d for rqs", __LINE__, r);
				goto out;
			}

			if (dflag & DKERNEL) {
				unpin_and_detach(Sbp);
			} else {
				unpin_and_detach(Sbp);
			}
            actual_sense = l->senselen - Ubp->b_resid;
			
			Trace4(1, "%d SENSE: b_flags %lx b_error %d b_resid %ld",
				   __LINE__, (long)Sbp->b_flags, Sbp->b_error,
				   (long)Sbp->b_resid);
			Trace5(1, "%d: sv %x st %x gc %x as %x", __LINE__,
				   usl->status_validity, usl->scsi_status,
				   usl->adapter_status, usl->adap_q_status);
			if (usl->scsi_status || usl->adapter_status) {
				r = EIO;
			} else {
				rqvalid = 1;
			}
		}
    }
	
    if (usc->status_validity & SC_ADAPTER_ERROR) {
		sp->needresume = 1;
		Trace2(0, "%d: adapter error 0x%x", __LINE__,
			   usc->adapter_status);
		Ubp->b_flags |= B_ERROR;
		switch (usc->adapter_status) {
		case SC_NO_DEVICE_RESPONSE:
		case SC_HOST_IO_BUS_ERR:
		case SC_SCSI_BUS_FAULT:
		case SC_CMD_TIMEOUT:
		case SC_ADAPTER_HDW_FAILURE:
		case SC_ADAPTER_SFW_FAILURE:
		case SC_FUSE_OR_TERMINAL_PWR:
		case SC_SCSI_BUS_RESET:
		default:
			r = EIO;
			break;
		}
    }
	
    /*
     * Log errors through errsave function
     */
    if (usc->status_validity & (SC_SCSI_ERROR|SC_ADAPTER_ERROR)) {
		struct sc_error_log_df log;
		
		bzero(&log, sizeof (log));
		/*
		 * All errors are 'temporary unknown driver error'
		 */
		log.error_id = ERRID_SCSI_ERR6;
		strncpy(log.resource_name, sp->devname, ERR_NAMESIZE);
		bcopy(&usc->scsi_command, &log.scsi_command, sizeof (struct scsi));
		log.status_validity = usc->status_validity;
		log.scsi_status = usc->scsi_status;
		log.general_card_status	= usc->adapter_status;
		if (rqvalid) {
			int amt;
			if (l->senselen > 128)
				amt = 128;
			else
				amt = l->senselen;
			(void) copyin(l->sense_buf, log.req_sense_data, amt);
		}
		errsave((struct err_rec*)&log, sizeof (struct sc_error_log_df));
    }
	
    if (dflag & DKERNEL) {
		*l->statusp = sbyte;
    } else {
		r2 = copyout(&sbyte, l->statusp, 1);
		if (r2 != 0) {
			if (r == 0)
				r = r2;
			goto out;
		}
    }
 out:
    if (l->datalen) {
		if (upin == 1) {
			unpin_and_detach(Ubp);
		}
    }
    l->datalen  = actual_data;
    l->senselen = actual_sense;

    Trace2(1, "%d: returning %d", __LINE__, r);

    return (r);
}

static int
make_rqs(gsc_softc_t * sp, char albits, char *uaddr, int ulen, int isk)
{
#ifdef FCP
    struct scsi_buf *usl;
#else
    struct sc_buf *usl;
#endif
    struct buf *Sbp;
    int err, upin;
	
    if (ulen > 255)
		ulen = 255;
    upin = err = 0;
    usl = &sp->rqsbuf;
    Sbp = &usl->bufstruct;
    bzero(usl, sizeof (*usl));
	
    Sbp->b_un.b_addr = uaddr;
    Sbp->b_resid = Sbp->b_bcount = ulen;
    Sbp->b_iodone = gscdd_intr;
    Sbp->b_dev = DEV32TODEV(sp->dev);
    Sbp->b_flags = B_BUSY | B_READ | B_MPSAFE;
    Sbp->b_event = EVENT_NULL;
    Sbp->b_xmemd.aspace_id = XMEM_INVAL;
	
    if (isk)
		err = pin_and_attach(Sbp, UIO_SYSSPACE);
    else
		err = pin_and_attach(Sbp, UIO_USERSPACE);
    if (err)
		goto out;
    upin = 1;

    err = 0;
	
#ifdef FCP
    usl->lun_id = sp->lun_id; /* Setting lun in sc_buf structure */
    usl->scsi_id = sp->scsi_id;	
#else
    usl->lun = sp->lun_id; /* Setting lun in sc_buf structure */
    usl->scsi_command.scsi_id = sp->scsi_id;
    usl->scsi_command.scsi_cmd.lun = albits & 0xe0; /* only copy the lun bits */
#endif
    usl->scsi_command.scsi_length = 6;
    usl->scsi_command.scsi_cmd.scsi_op_code = 0x3;
    usl->scsi_command.scsi_cmd.scsi_bytes[2+SCSICMD_SHIFT] = ulen;
    usl->timeout_value = 2;
    usl->flags = SC_RESUME;
    sp->needresume = 0;
	
 out:
    if (err) {
		if (upin == 1) {
			unpin_and_detach(Sbp);
		}
    }
    return (err);
}

void
gscdd_intr(struct buf * bp)
{
    int lv;
    gsc_softc_t *sp = &softinfo[((gsc_buf_t *)bp)->index];

    lv = disable_lock(INTIODONE, &sp->buf_lock);
    bp->b_flags |= B_DONE;
    unlock_enable(lv, &sp->buf_lock);
    e_wakeup(&bp->b_event);
}

/*
 * Pin the specified address range in user or system memory. Attach to
 * a user buffer for cross-memory operations.  
 */
static int pin_and_attach (struct buf *bufp, int space)
{
	int rc = 0;

	if (xmattach (bufp->b_baddr, bufp->b_bcount, &bufp->b_xmemd, space) != XMEM_SUCC) {
		rc = EFAULT;
		goto err_exit;
	}

	rc = xmempin (bufp->b_baddr, bufp->b_bcount, &bufp->b_xmemd);

	if (rc != PIN_SUCC) {
		xmdetach (&bufp->b_xmemd);
		rc = EFAULT;
	}

 err_exit:
	return rc;
}

/* 
 * Unpin the specified address range in user or system memory. Detach
 * to a user buffer for cross-memory operations.  
 */
static int unpin_and_detach (struct buf *bufp)
{
	int rc = 0;

	rc = xmemunpin (bufp->b_baddr, bufp->b_bcount, &bufp->b_xmemd);
	if (rc) {
		rc = EFAULT;
		goto err_exit;
	}

	rc = xmdetach (&bufp->b_xmemd);

	if (rc != XMEM_SUCC) {
		rc = EFAULT;
		goto err_exit;
	}
	
 err_exit:
	return rc;
}

/*
 * mode: c
 * Local variables:
 * c-indent-level: 4
 * c-brace-imaginary-offset: 0
 * c-brace-offset: -4
 * c-argdecl-indent: 4
 * c-label-offset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * End:
 */
