/*
 * %W% %G%
 * $Id:$
 * Copyright (c) 1997 by Matthew Jacob
 * Copyright (c) 2001 by Frank Zago
 *
 *	This software is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU Library General Public
 *	License as published by the Free Software Foundation; version 2.
 *
 *	This software is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *	Library General Public License for more details.
 *
 *	You should have received a copy of the GNU Library General Public
 *	License along with this software; if not, write to the Free
 *	Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *	The author may be reached via electronic communications at
 *
 *		fzago@greshamstorage.com
 */

#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>

#include "gscdds.h"
#include "gsctest.h"


int main(int argc, char *argv[])
{
    int fd = -1;
	scmd_t scmd;
	char sb[32], iqd[4096], sbyte;
	char cdb_any[16];
	int i;
	int main_ret = 1;

	/* CDROM characteristics. */
	int block_length;
	int last_block;

	if (argc != 2) {
		fprintf(stderr, "Usage: gsctest <device>\n");		
		goto the_end;
	}

	if ((fd = open(argv[1], 0)) == -1) {
		perror("open");
		goto the_end;
	}

	scmd.cdb = cdb_any;


	/* Try an inquiry. */
	printf("TEST: inquiry\n");
	{
		char vendor[9];
		char product[17];
		char rev[5];

		MK_CDB_INQUIRY(scmd, 255);
		scmd.data_buf = iqd;
		scmd.datalen = 255;
		scmd.sense_buf = sb;
		scmd.senselen = sizeof (sb);
		scmd.statusp = &sbyte;
		scmd.timeval = 5;

		if (ioctl(fd, GSC_CMD, (caddr_t) &scmd) == -1) {
			perror("GSC_CMD INQUIRY");
			goto the_end;
		}

		if (sbyte != 0) {
			fprintf(stderr, "error: sense/ASC/ASCQ = %d/%02x/%02x\n", 
					sbyte, sb[12], sb[13]);
			goto the_end;
		}

		strncpy(vendor, &iqd[8], 8); 
		vendor[8] = 0;

		strncpy(product, &iqd[16], 16);
		product[16] = 0;

		strncpy(rev, &iqd[32], 4);
		rev[4] = 0;

		fprintf(stdout, "    type=%d, vendor=[%s], product=[%s], rev=[%s]\n",
				iqd[0] & 0x1f, vendor, product, rev);
	}		
	printf("Test OK\n\n");


	if ((iqd[0] & 0x1f) != 0x05) {
		fprintf(stderr, "Continuing this test requires a CD-ROM drive with a data CD only.\n");
		goto the_end;
	}


	/* Test unit ready. */
	printf("TEST: TEST UNIT READY\n");
	{
		MK_CDB_TUR(scmd);
		scmd.data_buf = NULL;
		scmd.datalen = 0;
		scmd.sense_buf = sb;
		scmd.senselen = sizeof (sb);
		scmd.statusp = &sbyte;
		scmd.timeval = 5;

		if (ioctl(fd, GSC_CMD, (caddr_t) &scmd) == -1) {
			perror("GSC_CMD TUR");
			goto the_end;
		}

		if (sbyte != 0) {
			fprintf(stderr, "error: sense/ASC/ASCQ = %d/%02x/%02x\n", 
					sbyte, sb[12], sb[13]);
			goto the_end;
		}
	}
	printf("Test OK\n\n");

	
	/* Try an invalid command and check the sense. */
	printf("TEST: invalid command\n");	
	{
		char cdb_bad[10] = { 0x37, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

		scmd.cdb = cdb_bad;
		scmd.cdblen = 10;
		scmd.data_buf = NULL;
		scmd.datalen = 0;
		scmd.sense_buf = sb;
		scmd.senselen = sizeof (sb);
		scmd.statusp = &sbyte;
		scmd.rw = 0;
		scmd.timeval = 5;

		if (ioctl(fd, GSC_CMD, (caddr_t) &scmd) < 0) {
			perror("GSC_CMD bad command");
			goto the_end;
		}

		if (sbyte != 2 || sb[12]!=0x20 || sb[13]!=0x00) {
			fprintf(stderr, "error, got sense/ASC-ASCQ %d/%02x%02x instead of 2/2000\n", sbyte, sb[12], sb[13]);
			goto the_end;
		}
	}
	printf("Test OK\n\n");	

	
	/* Get the cdrom size. */
	printf("TEST: READ CAPACITY\n");
	{
		MK_CDB_READ_CAPACITY(scmd);
		scmd.data_buf = iqd;
		scmd.datalen = 8;
		scmd.sense_buf = sb;
		scmd.senselen = sizeof (sb);
		scmd.statusp = &sbyte;
		scmd.timeval = 5;
		
		if (ioctl(fd, GSC_CMD, (caddr_t) &scmd) == -1) {
			perror("GSC_CMD READ CAPACITY");
			goto the_end;
		}

		if (sbyte != 0) {
			fprintf(stderr, "error: sense/ASC/ASCQ = %d/%02x/%02x\n", 
					sbyte, sb[12], sb[13]);
			goto the_end;
		}

		last_block = iqd[0] << 24 | iqd[1] << 16 | iqd[2] << 8 | iqd[3];
		block_length = iqd[4] << 24 | iqd[5] << 16 | iqd[6] << 8 | iqd[7];

		printf("    number of blocks = %d\n", last_block);
		printf("    block length     = %d\n", block_length);
	}
	printf("Test OK\n\n");


	/* Some sanity checking. I have only 2048 bytes/blocks CDROM, but
	 * other values are possible. */
	if (block_length != 2048) {
		fprintf(stderr, "Invalid block size\n");
		goto the_end;
	}

	if (last_block < 1000) {
		fprintf(stderr, "Not enough blocks on that CDROM\n");
		goto the_end;
	}
	

	/* Do some read to stress the device a little. */
	printf("TEST: READ(10).\n");
	{

		for (i=0; i<1000; i++) {
			MK_CDB_READ_10(scmd, 2*i, 2);
			scmd.data_buf = iqd;
			scmd.datalen = 2 * block_length;
			scmd.sense_buf = sb;
			scmd.senselen = sizeof (sb);
			scmd.statusp = &sbyte;
			scmd.timeval = 5;
	
			if (ioctl(fd, GSC_CMD, (caddr_t) &scmd) == -1) {
				perror("GSC_CMD READ(10)");
				goto the_end;
			}

			if (sbyte != 0) {
				fprintf(stderr, "error: sense/ASC/ASCQ = %d/%02x/%02x\n", 
						sbyte, sb[12], sb[13]);
				goto the_end;
			}

		}
	}
	printf("Test OK\n\n");

	main_ret = 0;

 the_end:
	if (fd != -1)
		close(fd);

    return (main_ret);			/* 0 == no errors */
}

/*
 * mode: c
 * Local variables:
 * c-indent-level: 4
 * c-brace-imaginary-offset: 0
 * c-brace-offset: -4
 * c-argdecl-indent: 4
 * c-label-offset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * End:
 */
